import 'dart:math';

import '../extensions/extensions.dart';
import '../renderer/actor_member.dart';
import '../renderer/anchor.dart';
import '../renderer/matrix_transformer.dart';

class RectPoints implements ActorMember {
  @override
  final Vector2 scale = Vector2.all(1);

  Vector2 topLeft;
  Vector2 topRight;
  Vector2 bottomLeft;
  Vector2 bottomRight;
  @override
  Anchor anchor;

  bool _cached = false;
  Vector2? _size;
  double? _angle;
  Vector2? _skew;
  Vector2? _position;

  RectPoints({
    required this.topLeft,
    required this.topRight,
    required this.bottomLeft,
    required this.bottomRight,
    this.anchor = Anchor.center,
  });

  void updateAnchor(Anchor anchor) {
    this.anchor = anchor;
    _cached = false;
  }

  @override
  Vector2 get skew {
    _calculate();
    return _skew!;
  }

  @override
  Vector2 get position {
    _calculate();
    return _position!;
  }

  @override
  double get angle {
    _calculate();
    return _angle!;
  }

  @override
  Vector2 get size {
    _calculate();
    return _size!;
  }

  MatrixTransform get transform {
    _calculate();

    return MatrixTransform(
      skew: _skew,
      position: _position,
      scale: scale,
      offset: -Vector2(anchor.x * size.x, anchor.y * size.y),
      angle: _angle,
    );
  }

  void _calculate() {
    if (_cached) return;

    final size = Vector2(topRight.distanceTo(topLeft), bottomRight.distanceTo(topRight));
    final horizontalLineAngle = atan2(topRight.y - topLeft.y, topRight.x - topLeft.x);
    final verticalLineAngle = atan2(bottomLeft.y - topLeft.y, bottomLeft.x - topLeft.x);

    double angle = horizontalLineAngle;
    if (angle < 0) {
      angle += pi * 2;
    }

    double intersectionAngle = (horizontalLineAngle - verticalLineAngle).abs();
    if (intersectionAngle >= pi) {
      intersectionAngle = intersectionAngle - pi;
    }
    if (horizontalLineAngle >= 0 && verticalLineAngle < 0) {
      intersectionAngle = pi - intersectionAngle;
    }
    final skewAngle = pi / 2 - intersectionAngle;

    if (skewAngle != 0) {
      size.y *= cos(skewAngle.abs());
    }

    final leftTopPosition = (topLeft + bottomRight) / 2 - size / 2;
    final offset = Vector2(anchor.x * size.x, anchor.y * size.y);

    _position = leftTopPosition + offset;
    _size = size;
    _skew = Vector2(skewAngle, 0);
    _angle = angle;
    _cached = true;
  }
}
